from django.utils import timezone

from marshmallow import ValidationError, fields

from . import exceptions, schema


NOT_SPECIFIED = object()
OMIT = object()


class Field(fields.Field):
    """Custom field only to be used when parsing `query_params`.

    :param str param_name: Optional field that defines the name of the query parameter,
     it's set as optional because parsing lists do not require a name.
    :param any default: Optional
    :param bool many: Optional
    """

    @classmethod
    def __new__(cls, *args, **kwargs):
        if issubclass(cls, Nested):
            return super().__new__(cls)

        elif kwargs.pop("many", False):
            return _List(cls_or_instance=cls(**kwargs), **kwargs)

        return super().__new__(cls)

    def __init__(
        self,
        param_name=None,
        default=NOT_SPECIFIED,
        allow_none=False,
        validate=None,
        many=NOT_SPECIFIED,
        **kwargs,
    ):

        kwargs.update(
            {
                "allow_none": allow_none,
                "required": default is NOT_SPECIFIED,
                "validate": validate,
            }
        )

        if isinstance(self, Nested):
            kwargs["many"] = many
        elif many is not NOT_SPECIFIED:
            raise TypeError(
                "many should never be passed into __init__, it's only there to make mypy happy"
            )

        if param_name is not None:
            kwargs["data_key"] = param_name

        if default is not NOT_SPECIFIED and default is not OMIT:
            kwargs["missing"] = default

        super().__init__(**kwargs)


class Bool(fields.Boolean, Field):
    """Custom boolean field that supports validation."""


class DateTime(fields.DateTime, Field):
    """Custom datetime field that supports validation."""

    def _deserialize(self, value, attr, data, **kwargs):
        datetime = super()._deserialize(value, attr, data, **kwargs)
        if timezone.is_naive(datetime):
            raise exceptions.ValidationError(f"Timezone must be specified for {attr}.")

        return datetime


class Date(fields.Date, Field):
    """Custom date field that supports validation."""


class Int(fields.Integer, Field):
    """Custom int field that supports validation."""

    def __init__(self, *args, min_val=None, max_val=None, **kwargs):
        self.min_val = min_val
        self.max_val = max_val

        super().__init__(*args, **kwargs)

    def _validate_size(self, val):
        return (self.max_val is None or val <= self.max_val) and (
            self.min_val is None or val >= self.min_val
        )

    def _deserialize(self, value, attr, data, **kwargs):
        ret = super()._deserialize(value, attr, data, **kwargs)

        if not self._validate_size(ret):
            raise exceptions.ValidationError(f"Invalid interger value for {attr}")

        return ret


class String(fields.String, Field):
    """Custom str field that supports validation."""

    def __init__(
        self, *args, allow_empty=False, max_length=None, min_length=None, **kwargs
    ):
        self.allow_empty = allow_empty

        self.max_length = max_length
        self.min_length = min_length

        super().__init__(*args, **kwargs)

    def _validate_length(self, val):
        if self.allow_empty and val == "":
            return True

        return (self.max_length is None or len(val) <= self.max_length) and (
            self.min_length is None or len(val) >= self.min_length
        )

    def _deserialize(self, value, attr, data, **kwargs):
        ret = super()._deserialize(value, attr, data, **kwargs)

        if not self.allow_empty and ret == "":
            raise exceptions.ValidationError(f"String field cannot be empty for {attr}")

        if not self._validate_length(ret):
            raise exceptions.ValidationError(f"Invalid string length for {attr}")

        return ret


class Enum(fields.String, Field):
    def __init__(self, *args, choices=None, **kwargs):
        if choices is None:
            raise TypeError("choices must be specified")

        self.choices = choices

        super().__init__(*args, **kwargs)

    def _validate_choices(self, val):
        return val in self.choices

    def _deserialize(self, value, attr, data, **kwargs):
        ret = super()._deserialize(value, attr, data, **kwargs)

        if not self._validate_choices(ret):
            raise exceptions.ValidationError(f"Invalid choice for {attr}")

        return ret


class _List(fields.List, Field):
    """Custom list field that supports validation."""


class UUID(fields.UUID, Field):
    """UUID field."""


class Float(fields.Float, Field):
    """Float field."""


class Decimal(fields.Decimal, Field):
    """Decimal field."""

    def __init__(self, *args, max_digits=None, **kwargs):
        self.max_digits = max_digits

        super().__init__(*args, **kwargs)

    def _deserialize(self, *args, **kwargs):
        result = super()._deserialize(*args, **kwargs)

        if not self.max_digits:
            return result

        if self.max_digits and False:
            raise exceptions.ValidationError(
                f"Decimal field cannot be longer than {self.max_digits}"
            )

        return self.validate_precision(result)

    def validate_precision(self, value):
        """Validate the precision of the decimal field.

        # Simplified from: rest_framework.fields.Decimal.validate_precision
        """
        sign, digittuple, exponent = value.as_tuple()

        if exponent >= 0:
            # 1234500.0
            total_digits = len(digittuple) + exponent

        elif len(digittuple) > abs(exponent):
            # 123.45
            total_digits = len(digittuple)

        else:
            # 0.001234
            total_digits = abs(exponent)

        error_messages = []
        if self.max_digits is not None and total_digits > self.max_digits:
            error_messages.append(f"Digits cannot be longer than {self.max_digits}")

        if error_messages:
            errors = ", ".join(error_messages)
            raise exceptions.ValidationError(f"Invalid decimal field: {errors}")

        return value


class Time(fields.Time, Field):
    """Time field."""


class Email(fields.Email, Field):
    pass


class Nested(fields.Nested, Field):
    """Custom nested field."""

    def __init__(self, spec_or_schema, *args, **kwargs):
        self._needs_schema = False
        field_schema = schema.Schema

        if isinstance(spec_or_schema, dict):
            self.fields = spec_or_schema
            self._needs_schema = True
        else:
            self.fields = spec_or_schema().fields
            field_schema = spec_or_schema

        super().__init__(field_schema, *args, **kwargs)

    def _load(self, value, data, partial=None):
        try:
            valid_data = self.schema.load(value)
        except ValidationError as exc:
            raise ValidationError(exc.messages, valid_data=exc.valid_data)
        return valid_data

    def needs_schema(self):
        """Returns true if the field requires a schema to be set."""
        return self._needs_schema

    def bind_schema(self, nested_schema):
        """Bind the schema to the field."""
        if not self._needs_schema:
            return

        self.nested = nested_schema
        self.nested.unknown = nested_schema.unknown
        self.nested.many = self.many

        # Reset the internal schema
        self.__schema = None


class Dict(fields.Dict, Field):
    pass



from marshmallow import ValidationError as BaseValidationError


class ImmediateResponseError(Exception):
    def __init__(self, response):
        self.response = response


class ValidationError(BaseValidationError):
    """Base error for all of query param validation/parsing errors."""



import json

from django.test import RequestFactory, SimpleTestCase

from mbq.api_tools import fields, responses
from mbq.api_tools.exceptions import ImmediateResponseError
from mbq.api_tools.views import view


class TruePermission:
    def has_permission(self, *args, **kwargs):
        return True


class ErrorHandlingTests(SimpleTestCase):
    status_codes_to_responses = {
        400: responses.ClientErrorResponse,
        403: responses.UnauthorizedResponse,
        404: responses.NotFoundResponse,
        405: responses.MethodNotAllowedResponse,
        422: responses.ServerValidationErrorResponse,
    }

    def test_error_responses(self):
        @view(permissions=[TruePermission], params={"status_code": fields.Int()})
        def view_func(request):
            response_class = self.status_codes_to_responses[request.params.status_code]

            return response_class("unique_error_code", "here are the details")

        for status_code in self.status_codes_to_responses:
            request = RequestFactory().get("/some-url", {"status_code": status_code})

            response = view_func(request)

            self.assertEqual(response.status_code, status_code)
            self.assertEqual(response.data["error_code"], "unique_error_code")
            self.assertEqual(response.data["detail"], "here are the details")

    def test_immediate_response_exception(self):
        @view(permissions=[TruePermission])
        def view_func(request):
            raise ImmediateResponseError(
                responses.ClientErrorResponse(
                    "unique_error_code", "here are the details"
                )
            )

        request = RequestFactory().get("/some-url")
        response = view_func(request)

        self.assertEqual(response.status_code, 400)
        self.assertEqual(response.data["error_code"], "unique_error_code")
        self.assertEqual(response.data["detail"], "here are the details")


class UnknownFieldsTests(SimpleTestCase):
    def test_unknown_payload_field_settings(self):
        @view(
            "POST",
            permissions=[TruePermission],
            payload={"foo": fields.String(default=None)},
        )
        def view_func(request):
            return responses.DetailResponse()

        factory = RequestFactory()

        request = factory.post(
            "/some-url", json.dumps({"bar": "fizz"}), content_type="application/json"
        )

        with self.settings(API_TOOLS={}):
            response = view_func(request)

        self.assertEqual(response.status_code, 400)

        with self.settings(API_TOOLS={"UNKNOWN_PAYLOAD_FIELDS": "exclude"}):
            response = view_func(request)

        self.assertEqual(response.status_code, 200)

    def test_unknown_param_field_settings(self):
        @view(permissions=[TruePermission], params={"foo": fields.String(default=None)})
        def view_func(request):
            return responses.DetailResponse()

        factory = RequestFactory()

        request = factory.get(
            "/some-url", {"bar": "fizz"}, content_type="application/json"
        )

        with self.settings(API_TOOLS={}):
            response = view_func(request)

        self.assertEqual(response.status_code, 400)

        with self.settings(API_TOOLS={"UNKNOWN_PARAM_FIELDS": "exclude"}):
            response = view_func(request)

        self.assertEqual(response.status_code, 200)

    def test_unknown_override(self):
        @view(
            permissions=[TruePermission],
            params={"foo": fields.String(default=None)},
            on_unknown_field="exclude",
        )
        def view_func(request):
            return responses.DetailResponse()

        factory = RequestFactory()

        request = factory.get(
            "/some-url", {"bar": "fizz"}, content_type="application/json"
        )

        with self.settings(API_TOOLS={"UNKNOWN_PARAM_FIELDS": "raise"}):
            response = view_func(request)

        self.assertEqual(response.status_code, 200)



from django.test import SimpleTestCase

from mbq.api_tools import exceptions
from mbq.api_tools.pagination import paginate_queryset
from mbq.api_tools.schema import Pagination


class PaginationTests(SimpleTestCase):
    def test_pagination_structure(self):
        data = [{"id": idx} for idx in range(100)]
        url = "/test-pagination"
        paginator = Pagination(page=2, page_size=25)

        data, pagination_data = paginate_queryset(data, url, paginator)

        self.assertEqual(pagination_data["page"], 2)
        self.assertEqual(pagination_data["page_size"], 25)
        self.assertEqual(pagination_data["total_objects"], 100)
        self.assertEqual(pagination_data["num_pages"], 4)
        self.assertEqual(pagination_data["next_page"], "/test-pagination?page=3")
        self.assertEqual(pagination_data["previous_page"], "/test-pagination?page=1")
        self.assertEqual(len(data), 25)
        self.assertEqual(data[0]["id"], 25)

    def test_error_handling(self):
        data = [{"id": idx} for idx in range(10)]
        url = "/test-pagination"
        paginator = Pagination(page=3, page_size=25)

        with self.assertRaises(exceptions.ImmediateResponseError):
            paginate_queryset(data, url, paginator)

    def test_empty_results(self):
        url = "/test-pagination"
        paginator = Pagination(page=1, page_size=25)

        data, pagination_data = paginate_queryset([], url, paginator)

        self.assertEqual(pagination_data["page"], 1)
        self.assertEqual(pagination_data["page_size"], 25)
        self.assertEqual(pagination_data["total_objects"], 0)
        self.assertEqual(pagination_data["num_pages"], 1)
        self.assertEqual(pagination_data["next_page"], None)
        self.assertEqual(pagination_data["previous_page"], None)
        self.assertEqual(len(data), 0)



from unittest.mock import patch


def _disable_authorization(test_func):
    def wrapper(*args, **kwargs):
        with patch(
            "mbq.api_tools.views.ViewFunction._perform_authorization", return_value=None
        ):
            test_func(*args, **kwargs)

    return wrapper


def disable_authorization(func_or_cls):
    if isinstance(func_or_cls, type):
        for attr in dir(func_or_cls):
            if attr.startswith("test_"):
                setattr(
                    func_or_cls,
                    attr,
                    _disable_authorization(getattr(func_or_cls, attr)),
                )
        return func_or_cls
    else:
        return _disable_authorization(func_or_cls)
